@@ -70,6 +70,17 @@ class Driver {
 				e.Merge ();
 				
 				return 0;
+			
+			case "--local-edit":
+				if (i+1 == args.Length) {
+					Console.WriteLine ("--local-edit requires an argument");
+					return 1; 
+				}
+				EcmaUncompiledHelpSource hs = new EcmaUncompiledHelpSource(args[i+1]);
+				RootTree.ExtraHelpSources.Add(hs.Name + " (Local Edit)", hs);
+				i++;
+				break;
+				
 			default:
 				topic = args [i];
 				break;
@@ -82,8 +93,8 @@ class Driver {
 		Settings.RunningGUI = true;
 		Application.Init ();
 		Browser browser = new Browser ();
-
-		if (args.Length > 0)
+		
+		if (topic != null)
 			browser.LoadUrl (topic);
 		Application.Run ();
 		return 0;
