@@ -470,6 +470,11 @@ public class EcmaHelpSource : HelpSource {
 		ExtObject = new ExtensionObject (this);
 	}
 
+	public EcmaHelpSource ()
+	{
+		ExtObject = new ExtensionObject (this);
+	}
+
 	public override string GetText (string url, out Node match_node)
 	{
 		match_node = null;
@@ -532,7 +537,6 @@ public class EcmaHelpSource : HelpSource {
 			isoperator = true;
 			membername = "Conversion";
 			member = argtypes[0] + " to " + argtypes[1];
-			Console.WriteLine(member);
 		} else if (membername.StartsWith("op_")) {
 			isoperator = true;
 			membername = membername.Substring(3);
@@ -665,14 +669,14 @@ public class EcmaHelpSource : HelpSource {
 		return GetTextFromUrl (url);
 	}
 
-	public string GetIdFromUrl (string prefix, string ns, string type)
+	public virtual string GetIdFromUrl (string prefix, string ns, string type)
 	{
 		Node tmp_node = new Node (Tree, "", "");
 		string url = GetUrlForType (prefix, ns, type, null, out tmp_node);
 		if (url == null) return null;
 		return GetFile (url.Substring (5), out url);
 	}
-
+	
 	public string GetUrlForType (string prefix, string ns, string type, string member, out Node match_node)
 	{
 		if (!prefix.EndsWith(":")) throw new ArgumentException("prefix");
@@ -691,10 +695,21 @@ public class EcmaHelpSource : HelpSource {
 			
 			foreach (Node type_node in ns_node.Nodes){
 				string element = type_node.Element;
-				int pidx = element.IndexOf ("#");
-				int sidx = element.IndexOf ("/");
 				
-				string cname = element.Substring (pidx + 1, sidx-pidx-1);
+				string cname;
+				if (element.StartsWith("T:")) {
+					cname = element.Substring(2);
+					int pidx = cname.LastIndexOf (".");
+					cname = cname.Substring(pidx+1);
+					pidx = cname.LastIndexOf ("/");
+					if (pidx != -1)
+						cname = cname.Substring(0, pidx);
+					cname = cname.Replace("+", ".");
+				} else {				
+					int pidx = element.IndexOf ("#");
+					int sidx = element.IndexOf ("/");
+					cname = element.Substring (pidx + 1, sidx-pidx-1);
+				}
 				
 				//Console.WriteLine ("t:{0} cn:{1} p:{2}", type, cname, prefix);
 
@@ -723,7 +738,9 @@ public class EcmaHelpSource : HelpSource {
 							}
 						}
 						
-						return type_node.URL + type.Substring (p + 1);
+						string ret = type_node.URL;
+						if (!ret.EndsWith("/")) ret += "/";
+						return ret + type.Substring (p + 1);
 					}
 				}
 			}
@@ -735,6 +752,13 @@ public class EcmaHelpSource : HelpSource {
 	
 	public override string GetNodeXPath (XPathNavigator n)
 	{
+		if (n.Matches ("/Type/Docs/param")) {
+			string type_name = (string) n.Evaluate ("string (ancestor::Type/@FullName)");
+			string param_name = (string) n.Evaluate ("string (@name)");
+			
+			return String.Format ("/Type [@FullName = '{0}']/Docs/param[@name='{1}']", type_name, param_name);
+		}
+
 		if (n.Matches ("/Type/Docs/*")) {
 			string type_name = (string) n.Evaluate ("string (ancestor::Type/@FullName)");
 			
@@ -776,6 +800,9 @@ public class EcmaHelpSource : HelpSource {
 		return base.GetNodeXPath (n);
 	}
 
+	protected virtual XmlReader GetNamespaceDocument (string ns) {
+		return GetHelpXml ("xml.summary." + ns);
+	}
 
 	public override string RenderNamespaceLookup (string nsurl, out Node match_node)
 	{
@@ -786,7 +813,7 @@ public class EcmaHelpSource : HelpSource {
 			match_node = ns_node;
 			string ns_name = nsurl.Substring (2);
 			
-			XmlReader summary = GetHelpXml ("xml.summary." + ns_name);
+			XmlReader summary = GetNamespaceDocument (ns_name);
 			if (summary == null)
 				return null;
 
@@ -811,29 +838,23 @@ public class EcmaHelpSource : HelpSource {
 	}
 
 	//
-	// Returns the XmlDocument from the given url, and fills in `file' and `rest'
+	// Returns the XmlDocument from the given url, and fills in `rest'
 	//
-	XmlDocument GetXmlFromUrl (string url, out string file, out string rest)
-	{
-		//
+	protected virtual XmlDocument GetXmlFromUrl(string url, out string rest) {
 		// Remove ecma:
-		//
 		url = url.Substring (5);
-		file = GetFile (url, out rest);
+		string file = GetFile (url, out rest);
 
 		// Console.WriteLine ("Got [{0}] and [{1}]", file, rest);
-		XmlDocument doc = GetHelpXmlWithChanges (file);
-		if (doc == null)
-			return null;
-
-		return doc;
+		return GetHelpXmlWithChanges (file);
 	}
 	
 	string GetTextFromUrl (string url)
 	{
-		string file, rest;
+		string rest, rest2;
+		Node node;
 
-		XmlDocument doc = GetXmlFromUrl (url, out file, out rest);
+		XmlDocument doc = GetXmlFromUrl (url, out rest);
 		
 		// Load base-type information so the stylesheet can draw the inheritance
 		// tree and (optionally) include inherited members in the members list.
@@ -852,11 +873,10 @@ public class EcmaHelpSource : HelpSource {
 			int dot = basetype.InnerText.LastIndexOf('.');
 			string ns = basetype.InnerText.Substring(0, dot == -1 ? 0 : dot);
 			string n = basetype.InnerText.Substring(dot == -1 ? 0 : dot+1);
-			string basetypexml = GetIdFromUrl("T:", ns, n);
+			string basetypeurl = GetUrlForType("T:", ns, n, null, out node);
 			XmlDocument basetypedoc = null;
-
-			if (basetypexml != null)
-				basetypedoc = GetHelpXmlWithChanges (basetypexml);
+			if (basetypeurl != null)
+				basetypedoc = GetXmlFromUrl (basetypeurl, out rest2);
 			if (basetypedoc == null) {
 				inheritancenode.SetAttribute("Incomplete", "1");
 				break;
@@ -978,19 +998,18 @@ public class EcmaHelpSource : HelpSource {
 		EnsureTransform ();
 		
 		StringWriter output = new StringWriter ();
-		ecma_transform.Transform (ecma_xml, args, output);
+		ecma_transform.Transform (ecma_xml, args, output, null);
 		return output.ToString ();
 	}
 	
 	static void Htmlize (IXPathNavigable ecma_xml, XsltArgumentList args, XmlWriter w)
 	{
-
 		EnsureTransform ();
 		
 		if (ecma_xml == null)
 			return;
 
-		ecma_transform.Transform (ecma_xml, args, w);
+		ecma_transform.Transform (ecma_xml, args, w, null);
 	}
 	
 	static void EnsureTransform ()
@@ -1001,7 +1020,7 @@ public class EcmaHelpSource : HelpSource {
 			Stream stream = assembly.GetManifestResourceStream ("mono-ecma.xsl");
 
 			XmlReader xml_reader = new XmlTextReader (stream);
-			ecma_transform.Load (xml_reader);
+			ecma_transform.Load (xml_reader, null, null);
 		}
 	}
 
@@ -1022,7 +1041,7 @@ public class EcmaHelpSource : HelpSource {
 			return(Mono.Utilities.Colorizer.Colorize(code,lang));
 		}
 		// Used by stylesheet to nicely reformat the <see cref=> tags. 
-		public string MakeNiceSignature(string sig)
+		public string MakeNiceSignature(string sig, string contexttype)
 		{
 			if (sig.Length < 3)
 				return sig;
@@ -1031,28 +1050,47 @@ public class EcmaHelpSource : HelpSource {
 
 			char s = sig[0];
 			sig = sig.Substring(2);
-
+			
 			switch (s) {
 				case 'N': return sig;
-				case 'T': return ShortTypeName(sig);
+				case 'T': return ShortTypeName(sig, contexttype);
 
 				case 'C': case 'M': case 'P': case 'F': case 'E':
-					int paren = sig.IndexOf("(");
-					if (paren > 0) {
-						string name = sig.Substring(0, paren);
-						string[] args = sig.Substring(paren, sig.Length-paren-1).Split(',');
-						
-						// avoid (() on no args
-						if (args.Length == 1 && args[0] == "(")
-							return ShortTypeName(sig);
-							
+					string type, mem, arg;
+					
+					// Get arguments
+					int paren;
+					if (s == 'C' || s == 'M')
+						paren = sig.IndexOf("(");
+					else if (s == 'P')
+						paren = sig.IndexOf("[");
+					else
+						paren = 0;
+					
+					if (paren > 0 && paren < sig.Length-1) {
+						string[] args = sig.Substring(paren+1, sig.Length-paren-2).Split(',');						
 						for (int i = 0; i < args.Length; i++)
-							args[i] = ShortTypeName(args[i]);
-						return ShortTypeName(name) + "(" + String.Join(", ", args) + ")";
+							args[i] = ShortTypeName(args[i], contexttype);
+						arg = "(" + String.Join(", ", args) + ")";
+						sig = sig.Substring(0, paren); 
 					} else {
-						return ShortTypeName(sig);
+						arg = "";
 					}
 
+					// Get type and member names
+					int dot = sig.LastIndexOf(".");
+					if (s == 'C' || dot <= 0 || dot == sig.Length-1) {
+						mem = "";
+						type = sig;
+					} else {
+						type = sig.Substring(0, dot);
+						mem = sig.Substring(dot);
+					}
+						
+					type = ShortTypeName(type, contexttype);
+					
+					return type + mem + arg;
+
 				default:
 					return sig;
 			}
@@ -1066,10 +1104,26 @@ public class EcmaHelpSource : HelpSource {
 			return "";
 		}
 
-		private static string ShortTypeName(string name)
+		public string EditUrlNamespace (string ns, string section)
 		{
-			//if (name.LastIndexOf(".") > 0) name = name.Substring(name.LastIndexOf(".")+1);
-			return name;
+			if (hs is EcmaUncompiledHelpSource)
+				return "edit:file:" + Path.Combine(((EcmaUncompiledHelpSource)hs).BasePath, ns + ".xml") + "@/Namespace/Docs/" + section; 
+			return "";
+		}
+
+		private static string ShortTypeName(string name, string contexttype)
+		{
+			int dot = contexttype.LastIndexOf(".");
+			if (dot < 0) return name;
+			string contextns = contexttype.Substring(0, dot+1);
+
+			if (name == contexttype)
+				return name.Substring(dot+1);
+			
+			if (name.StartsWith(contextns))
+				return name.Substring(contextns.Length);
+			
+			return name.Replace("+", ".");
 		}
 
 
@@ -1368,8 +1422,8 @@ public class EcmaHelpSource : HelpSource {
 					index_maker.Add (full + " " + doc_tag, full, url);
 
 					// Now, pull the values.
-					string file, rest;
-					XmlDocument x = GetXmlFromUrl (type_node.URL, out file, out rest);
+					string rest;
+					XmlDocument x = GetXmlFromUrl (type_node.URL, out rest);
 					if (x == null)
 						continue;
 					
@@ -1391,4 +1445,170 @@ public class EcmaHelpSource : HelpSource {
 		}
 	}
 }
+
+public class EcmaUncompiledHelpSource : EcmaHelpSource {
+	readonly DirectoryInfo basedir;
+	readonly XmlDocument basedoc;
+	
+	public new readonly string Name, BasePath;
+	
+	public EcmaUncompiledHelpSource (string base_file) : base ()
+	{
+		Console.Error.WriteLine("Loading uncompiled help from " + base_file);
+		
+		basedir = new DirectoryInfo(base_file);
+		BasePath = basedir.FullName;
+		
+		basedoc = new XmlDocument();
+		basedoc.Load(Path.Combine(basedir.FullName, "index.xml"));
+		
+		Name = basedoc.SelectSingleNode("Overview/Assembly").InnerText;
+		
+		bool has_content = false;
+		
+		foreach (XmlElement ns in basedoc.SelectNodes("Overview/Types/Namespace")) {
+			has_content = true;
+			Node nsnode = Tree.CreateNode(ns.GetAttribute("Name"), "N:" + ns.GetAttribute("Name"));
+			
+			bool has_types = false;
+			foreach (XmlElement t in ns.SelectNodes("Type")) {
+				has_types = true;
+				string typename = t.GetAttribute("Name").Replace("+", ".");
+				
+				string url = ns.GetAttribute("Name") + "." + t.GetAttribute("Name");
+				Node typenode = nsnode.CreateNode(typename, "T:" + url);				
+				Node typemembers = typenode.CreateNode("Members", "T:" + url + "/*");
+				
+				// Must load in each document to get the list of members...
+				XmlDocument typedoc = new XmlDocument();
+				typedoc.Load(Path.Combine(Path.Combine(basedir.FullName, ns.GetAttribute("Name")), t.GetAttribute("Name") + ".xml"));
+				
+				Hashtable groups = new Hashtable();
+				Hashtable groups_count = new Hashtable();
+				foreach (XmlElement member in typedoc.SelectNodes("Type/Members/Member")) {
+					string membername = member.GetAttribute("MemberName");
+					string membertype = member.SelectSingleNode("MemberType").InnerText;
+					
+					if (membertype == "Constructor")
+						membername = t.GetAttribute("Name");
+					if (membername.StartsWith("op_"))
+						membertype = "Operator";
+					
+					Node group;
+					if (groups.ContainsKey(membertype)) {
+						group = (Node)groups[membertype];
+					} else {
+						string membertypeplural = membertype + "s";
+						if (membertypeplural == "Propertys") membertypeplural = "Properties";
+						
+						group = typenode.CreateNode(membertypeplural, "T:" + url + "/" + membertype[0]);
+						groups[membertype] = group;
+						groups_count[membertype] = 0;
+					}
+					
+					if (membertype == "Constructor" || membertype == "Method" || 
+						(membertype == "Property" && member.SelectNodes("Parameters/Parameter").Count > 0)) {
+						membername = EcmaHelpSource.MakeSignature(member, membertype == "Constructor" ? membername : null);
+					} else if (membertype == "Operator") {
+						string dummy;
+						EcmaHelpSource.MakeOperatorSignature(member, out dummy, out membername);
+					}
+					
+					int index = (int)groups_count[membertype];
+					groups_count[membertype] = index + 1;
+					
+					group.CreateNode(membername, index.ToString());
+				}
+
+				foreach (Node group in groups.Values)
+					group.Sort();			
+			}
+			
+			if (has_types)
+				nsnode.Sort();
+		}
+		
+		if (has_content)
+			Tree.Sort();
+	}
+	
+	public override string GetIdFromUrl (string prefix, string ns, string type)
+	{
+		if (prefix != "T:")
+			throw new NotImplementedException();
+		return Path.Combine(Path.Combine(basedir.FullName, ns), type + ".xml");
+	}
+
+	protected override XmlDocument GetXmlFromUrl(string url, out string rest) {
+		// strip off the T:
+		url = url.Substring(2);
+		
+		int sidx = url.IndexOf("/");
+		if (sidx == -1) {
+			rest = "";
+		} else {
+			rest = url.Substring(sidx+1);
+			url = url.Substring(0, sidx);
+		}
+		
+		int pidx = url.LastIndexOf(".");
+		string ns = url.Substring(0, pidx);
+		string type = url.Substring(pidx+1).Replace(".", "+");
+		
+		string file = Path.Combine(Path.Combine(basedir.FullName, ns), type + ".xml");
+		if (!new FileInfo(file).Exists) return null;
+		
+		XmlDocument typedoc = new XmlDocument();
+		typedoc.Load(file);
+		return typedoc;
+	}
+	
+	public override string GetText (string url, out Node match_node) {
+		if (url == "root:") {
+			match_node = null;
+			return null;
+		}
+		return base.GetText(url, out match_node);
+	}
+	
+	protected override XmlReader GetNamespaceDocument (string ns) {
+		XmlDocument nsdoc = new XmlDocument();
+		nsdoc.Load(Path.Combine(basedir.FullName, ns + ".xml"));
+		
+		XmlDocument elements = new XmlDocument();
+		XmlElement docnode = elements.CreateElement("elements");
+		
+		foreach (XmlElement doc in nsdoc.SelectNodes("Namespace/Docs/*")) {
+			docnode.AppendChild(elements.ImportNode(doc, true));
+		}
+				
+		foreach (XmlElement t in basedoc.SelectNodes("Overview/Types/Namespace[@Name='" + ns + "']/Type")) {
+			XmlDocument typedoc = new XmlDocument();
+			typedoc.Load(Path.Combine(Path.Combine(basedir.FullName, ns), t.GetAttribute("Name") + ".xml"));
+			
+			string typekind;
+			switch (EcmaDoc.GetTypeKind(typedoc)) {
+			case "Class": typekind = "class"; break;
+			case "Enumeration": typekind = "enum"; break;
+			case "Structure": typekind = "struct"; break;
+			case "Delegate": typekind = "delegate"; break;
+			case "Interface": typekind = "interface"; break;
+			default: throw new InvalidOperationException();
+			}
+			
+			XmlElement typenode = elements.CreateElement(typekind);
+			typenode.SetAttribute("name", t.GetAttribute("Name"));
+			typenode.SetAttribute("fullname", ns + "." + t.GetAttribute("Name"));
+			typenode.AppendChild(elements.ImportNode(typedoc.SelectSingleNode("Type/Docs/summary"), true));
+			
+			docnode.AppendChild(typenode);
+		}
+
+		return new XmlNodeReader(docnode);
+	}
+	
 }
+
+}
+
+
